import enum
from typing import Optional, Annotated

from func_validator import MustBeMemberOf, validate_func_args

from geolysis.foundation import Shape, create_foundation
from geolysis.utils import AbstractStrEnum
from ._core import UltimateBearingCapacity
from ._hansen_ubc import HansenUltimateBearingCapacity
from ._terzaghi_ubc import (
    TerzaghiUBC4CircularFooting,
    TerzaghiUBC4RectangularFooting,
    TerzaghiUBC4SquareFooting,
    TerzaghiUBC4StripFooting,
)
from ._vesic_ubc import VesicUltimateBearingCapacity

__all__ = [
    "TerzaghiUBC4StripFooting",
    "TerzaghiUBC4CircularFooting",
    "TerzaghiUBC4RectangularFooting",
    "TerzaghiUBC4SquareFooting",
    "HansenUltimateBearingCapacity",
    "VesicUltimateBearingCapacity",
    "UBCType",
    "create_ubc_4_all_soil_types",
]


class UBCType(AbstractStrEnum):
    """Enumeration of available ultimate bearing capacity methods.

    Each member represents a different method for determining
    the ultimate bearing capacity of soil.
    """

    HANSEN = enum.auto()
    """Hansen's method for calculating ultimate bearing capacity."""

    TERZAGHI = enum.auto()
    """Terzaghi's method for calculating ultimate bearing capacity."""

    VESIC = enum.auto()
    """Vesic's method for calculating ultimate bearing capacity."""


@validate_func_args
def create_ubc_4_all_soil_types(
        friction_angle: float,
        cohesion: float,
        moist_unit_wgt: float,
        depth: float,
        width: float,
        length: Optional[float] = None,
        saturated_unit_wgt: float = 20.5,
        eccentricity: float = 0.0,
        ground_water_level: Optional[float] = None,
        load_angle: float = 0.0,
        apply_local_shear: bool = False,
        shape: Shape | str = "square",
        ubc_type: Annotated[UBCType | str, MustBeMemberOf(UBCType)] = "vesic",
) -> UltimateBearingCapacity:
    r"""A factory function that encapsulate the creation of ultimate
    bearing capacity.

    :param friction_angle: Internal angle of friction for general shear
                           failure (degree).

    :param cohesion: Cohesion of soil ($kPa$).
    :param moist_unit_wgt: Moist unit weight of soil ($kN/m^3$).
    :param depth: Depth of foundation (m).
    :param width: Width of foundation footing (m).
    :param length: Length of foundation footing (m).
    :param saturated_unit_wgt: Saturated unit weight of soil ($kN/m^3$).
    :param eccentricity: The deviation of the foundation load from the
                         center of gravity of the foundation footing.
    :param ground_water_level: Depth of water below ground level (m).
    :param load_angle: Inclination of the applied load with the  vertical
                       ($\alpha^{\circ}$).
    :param apply_local_shear: Indicate whether bearing capacity failure
                              is general or local shear failure.
    :param shape: Shape of foundation footing.
    :param ubc_type: Type of allowable bearing capacity calculation to
                     apply.

    :raises ValidationError: Raised if ubc_type is not supported.
    :raises ValidationError: Raised if an invalid footing shape is
                             provided.
    :raises ValueError: Raised when length is not provided for a
                        rectangular footing.
    """
    ubc_type = UBCType(ubc_type)

    # exception from create_foundation will automatically propagate
    # no need to catch and handle it.
    fnd_size = create_foundation(
        depth=depth,
        width=width,
        length=length,
        eccentricity=eccentricity,
        load_angle=load_angle,
        ground_water_level=ground_water_level,
        shape=shape,
    )

    ubc_class = _get_ultimate_bearing_capacity(ubc_type,
                                               fnd_size.footing_shape)

    return ubc_class(
        friction_angle=friction_angle,
        cohesion=cohesion,
        moist_unit_wgt=moist_unit_wgt,
        saturated_unit_wgt=saturated_unit_wgt,
        foundation_size=fnd_size,
        apply_local_shear=apply_local_shear,
    )


def _get_ultimate_bearing_capacity(ubc_type: UBCType, foundation_shape: Shape):
    ubc_classes = {
        UBCType.HANSEN: HansenUltimateBearingCapacity,
        UBCType.TERZAGHI: {
            Shape.STRIP: TerzaghiUBC4StripFooting,
            Shape.CIRCLE: TerzaghiUBC4CircularFooting,
            Shape.SQUARE: TerzaghiUBC4SquareFooting,
            Shape.RECTANGLE: TerzaghiUBC4RectangularFooting,
        },
        UBCType.VESIC: VesicUltimateBearingCapacity,
    }
    if ubc_type == UBCType.TERZAGHI:
        ubc_class = ubc_classes[ubc_type][foundation_shape]
    else:
        ubc_class = ubc_classes[ubc_type]
    return ubc_class



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Annotated, Optional

from func_validator import (
    validate_func_args,
    MustBeNonNegative,
    MustBePositive,
)

from geolysis.foundation import Foundation
from geolysis.utils import arctandeg, round_, tandeg


@dataclass(frozen=True, slots=True)
class UltimateBearingCapacityResult:
    ultimate_bearing_capacity: float
    allowable_bearing_capacity: float
    allowable_applied_load: float
    n_c: float
    n_q: float
    n_gamma: float
    s_c: float
    s_q: float
    s_gamma: float
    d_c: float
    d_q: float
    d_gamma: float
    i_c: float
    i_q: float
    i_gamma: float


class UltimateBearingCapacity(ABC):
    def __init__(
        self,
        friction_angle: float,
        cohesion: float,
        moist_unit_wgt: float,
        foundation_size: Foundation,
        saturated_unit_wgt: float = 20.5,
        apply_local_shear: bool = False,
        factor_of_safety: float = 3.0,
    ) -> None:
        r"""
        :param friction_angle: Internal angle of friction for general
                               shear failure (degrees).
        :param cohesion: Cohesion of soil ($kPa$).
        :param moist_unit_wgt: Moist unit weight of soil ($kN/m^3$).
        :param foundation_size: Size of the foundation.
        :param saturated_unit_wgt: Saturated unit weight of soil ($kN/m^3$).
        :param factor_of_safety: Factor of safety against bearing
                                 capacity failure. Added in v0.12.0.
        :param apply_local_shear: Indicate whether bearing capacity
                                  failure is general shear or local
                                  shear failure.
        """
        self.friction_angle = friction_angle
        self.cohesion = cohesion
        self.moist_unit_wgt = moist_unit_wgt
        self.foundation_size = foundation_size
        self.saturated_unit_wgt = saturated_unit_wgt
        self.factor_of_safety = factor_of_safety
        self.apply_local_shear = apply_local_shear

    @property
    def friction_angle(self) -> float:
        r"""Return friction angle for local shear in the case of local shear
        failure or general shear in the case of general shear failure.

        In the case of local shear failure:

        $$
        \phi' = \tan^{-1} \left(\frac{2}{3} \tan \phi\right)
        $$

        """
        if self.apply_local_shear:
            return arctandeg((2.0 / 3.0) * tandeg(self._friction_angle))
        return self._friction_angle

    @friction_angle.setter
    @validate_func_args
    def friction_angle(self, val: Annotated[float, MustBeNonNegative]):
        self._friction_angle = val

    @property
    def cohesion(self) -> float:
        r"""Return cohesion for local shear in the case of local shear failure
        or general shear in the case of general shear failure.

        In the case of local shear failure:

        $$C^{'} = \dfrac{2}{3} \cdot C$$
        """
        if self.apply_local_shear:
            return (2.0 / 3.0) * self._cohesion
        return self._cohesion

    @cohesion.setter
    @validate_func_args
    def cohesion(self, val: Annotated[float, MustBeNonNegative]):
        self._cohesion = val

    @property
    def moist_unit_wgt(self) -> float:
        """Moist unit weight of soil ($kN/m^3$)."""
        return self._moist_unit_wgt

    @moist_unit_wgt.setter
    @validate_func_args
    def moist_unit_wgt(self, val: Annotated[float, MustBePositive]):
        self._moist_unit_wgt = val

    @property
    def saturated_unit_wgt(self) -> float:
        """Saturated unit weight of soil ($kN/m^3$)."""
        return self._saturated_unit_wgt

    @saturated_unit_wgt.setter
    @validate_func_args
    def saturated_unit_wgt(self, val: Annotated[float, MustBePositive]):
        self._saturated_unit_wgt = val

    @property
    def load_angle(self):
        """Inclination of the applied load with the  vertical."""
        return self.foundation_size.load_angle

    @property
    def s_c(self) -> float:
        return 1.0

    @property
    def s_q(self) -> float:
        return 1.0

    @property
    def s_gamma(self) -> float:
        return 1.0

    @property
    def d_c(self) -> float:
        return 1.0

    @property
    def d_q(self) -> float:
        return 1.0

    @property
    def d_gamma(self) -> float:
        return 1.0

    @property
    def i_c(self) -> float:
        return 1.0

    @property
    def i_q(self) -> float:
        return 1.0

    @property
    def i_gamma(self) -> float:
        return 1.0

    def _cohesion_term(self, coef: float = 1.0) -> float:
        return coef * self.cohesion * self.n_c * self.s_c * self.d_c * self.i_c

    def _surcharge_term(self) -> float:
        depth = self.foundation_size.depth
        water_level = self.foundation_size.ground_water_level

        unit_wgt = self.moist_unit_wgt
        eop = unit_wgt * depth
        if water_level is not None:
            if water_level < depth:
                d_1 = water_level
                d_2 = depth - d_1
                unit_wgt = self.saturated_unit_wgt - 9.81
                eop = self.moist_unit_wgt * d_1 + unit_wgt * d_2
        return eop * self.n_q * self.s_q * self.d_q * self.i_q

    def _embedment_term(self, coef: float = 0.5) -> float:
        depth = self.foundation_size.depth
        width = self.foundation_size.effective_width
        water_level = self.foundation_size.ground_water_level

        unit_wgt = self.moist_unit_wgt

        if water_level is not None:
            wgt = self.saturated_unit_wgt - 9.81
            if water_level < depth:
                unit_wgt = wgt
            else:
                d = water_level - depth
                if d <= width:
                    unit_wgt = wgt + (d / width) * (self.moist_unit_wgt - wgt)

        return (
            coef
            * unit_wgt
            * width
            * self.n_gamma
            * self.s_gamma
            * self.d_gamma
            * self.i_gamma
        )

    def _bearing_capacity(self) -> float:
        return (
            self._cohesion_term(1.0)
            + self._surcharge_term()
            + self._embedment_term(0.5)
        )

    def bearing_capacity_results(self) -> UltimateBearingCapacityResult:
        """Return a dictionary of bearing capacity results with
        intermediate calculations.

        !!! info "Added in v0.11.0"

        """
        return UltimateBearingCapacityResult(
            ultimate_bearing_capacity=self.ultimate_bearing_capacity(),
            allowable_bearing_capacity=self.allowable_bearing_capacity(),
            allowable_applied_load=self.allowable_applied_load(),
            n_c=self.n_c,
            n_q=self.n_q,
            n_gamma=self.n_gamma,
            s_c=self.s_c,
            s_q=self.s_q,
            s_gamma=self.s_gamma,
            d_c=self.d_c,
            d_q=self.d_q,
            d_gamma=self.d_gamma,
            i_c=self.i_c,
            i_q=self.i_q,
            i_gamma=self.i_gamma,
        )

    @round_(ndigits=1)
    def ultimate_bearing_capacity(self) -> float:
        """Calculates the ultimate bearing capacity.

        !!! info "Added in v0.12.0"
        """
        return self._bearing_capacity()

    @round_(ndigits=1)
    def allowable_bearing_capacity(self) -> float:
        """Calculates the allowable bearing capacity.

        !!! info "Added in v0.12.0"
        """
        return self._bearing_capacity() / self.factor_of_safety

    @round_(ndigits=1)
    def allowable_applied_load(self) -> float:
        """Calculates the allowable applied load.

        !!! info "Added in v0.12.0"
        """
        area = self.foundation_size.foundation_area()
        return self.allowable_bearing_capacity() * area

    @property
    @abstractmethod
    def n_c(self) -> float: ...

    @property
    @abstractmethod
    def n_q(self) -> float: ...

    @property
    @abstractmethod
    def n_gamma(self) -> float: ...



import pytest

from geolysis.soil_classifier import (
    PSD,
    AtterbergLimits,
    create_aashto_classifier,
    create_uscs_classifier,
)


class TestAtterbergLimits:
    @pytest.mark.parametrize(
        ["liquid_lmt", "plastic_lmt", "expected"], [(25.0, 15.0, 10.0)]
    )
    def test_plasticity_idx(
        self, liquid_lmt: float, plastic_lmt: float, expected: float
    ):
        al = AtterbergLimits(liquid_limit=liquid_lmt, plastic_limit=plastic_lmt)
        assert al.plasticity_index == pytest.approx(expected)

    @pytest.mark.parametrize(
        ["liquid_lmt", "plastic_lmt", "nmc", "expected"], [(25.0, 15.0, 20.0, 50.0)]
    )
    def test_liquidity_idx(
        self, liquid_lmt: float, plastic_lmt: float, nmc: float, expected: float
    ):
        al = AtterbergLimits(liquid_limit=liquid_lmt, plastic_limit=plastic_lmt)
        assert al.liquidity_index(nmc=nmc) == pytest.approx(expected)

    @pytest.mark.parametrize(
        ["liquid_lmt", "plastic_lmt", "nmc", "expected"], [(25.0, 15.0, 20.0, 50.0)]
    )
    def test_consistency_idx(
        self, liquid_lmt: float, plastic_lmt: float, nmc: float, expected: float
    ):
        al = AtterbergLimits(liquid_limit=liquid_lmt, plastic_limit=plastic_lmt)
        assert al.consistency_index(nmc=nmc) == pytest.approx(expected)

    def test_errors(self):
        # Plastic limit is greater than liquid limit
        with pytest.raises(ValueError):
            AtterbergLimits(liquid_limit=15.0, plastic_limit=25.0)


class TestPSD:

    @pytest.mark.parametrize(
        ["fines", "sand", "d_10", "d_30", "d_60", "expected"],
        [(0.0, 0.0, 0.115, 0.53, 1.55, 13.48)],
    )
    def test_coeff_of_uniformity(self, fines, sand, d_10, d_30, d_60, expected):
        psd = PSD(fines=fines, sand=sand, d_10=d_10, d_30=d_30, d_60=d_60)
        assert psd.coeff_of_uniformity == pytest.approx(expected)

    @pytest.mark.parametrize(
        ["fines", "sand", "d_10", "d_30", "d_60", "expected"],
        [(0.0, 0.0, 0.115, 0.53, 1.55, 1.58)],
    )
    def test_coeff_of_curvature(self, fines, sand, d_10, d_30, d_60, expected):
        psd = PSD(fines=fines, sand=sand, d_10=d_10, d_30=d_30, d_60=d_60)
        assert psd.coeff_of_curvature == pytest.approx(expected)


class TestAASHTO:
    @pytest.mark.parametrize(
        ["al", "fines", "expected"],
        [
            ((30, 25), 10, "A-1-a(0)"),
            ((17.9, 14.5), 24.01, "A-1-b(0)"),
            ((0, 0), 9.5, "A-3(0)"),
            ((30.2, 23.9), 11.18, "A-2-4(0)"),
            ((30.1, 16.5), 18.38, "A-2-6(0)"),
            ((35, 28), 40.0, "A-4(1)"),
            ((48, 39), 40.0, "A-5(1)"),
            ((37.7, 23.8), 47.44, "A-6(4)"),
            ((61.7, 32.3), 52.09, "A-7-5(12)"),
            ((70.0, 38.0), 86, "A-7-5(20)"),
            ((52.6, 27.6), 45.8, "A-7-6(7)"),
            ((45.0, 16.0), 60, "A-7-6(13)"),
        ],
    )
    def test_aashto_with_grp_idx(self, al, fines, expected):
        asshto_clf = create_aashto_classifier(*al, fines=fines)
        assert asshto_clf.classify().symbol == expected

    @pytest.mark.parametrize(
        ["al", "fines", "expected"],
        [
            ((0.0, 0.0), 9.5, "A-3"),
            ((43.0, 35.0), 30.0, "A-2-5"),
            ((43.0, 25.0), 30.0, "A-2-7"),
            ((35.0, 28.0), 40.0, "A-4"),
        ],
    )
    def test_aashto_without_grp_idx(self, al, fines, expected):
        asshto_clf = create_aashto_classifier(*al, fines=fines)
        assert asshto_clf.classify().symbol_no_group_idx == expected


class TestUSCS:
    @pytest.mark.parametrize(
        ["al", "psd", "dist", "expected"],
        [
            ((30.8, 20.7), (10.29, 81.89), (0.07, 0.3, 0.8), "SW-SC"),
            ((24.4, 14.7), (9.77, 44.82), (0.06, 0.6, 7), "GP-GC"),
            ((49.5, 33.6), (6.93, 91.79), (0.153, 0.4, 1.2), "SP-SM"),
            ((30.33, 23.42), (8.93, 7.69), (0.15, 18, 44), "GP-GM"),
            ((35.32, 25.57), (9.70, 5.63), (0.06, 50, 55), "GP-GM"),
            ((26.17, 19.69), (12.00, 8.24), (0.07, 15, 52), "GP-GC"),
            ((30.59, 24.41), (9.87, 19.03), (0.07, 0.3, 0.8), "GW-GM"),
            ((32.78, 22.99), (3.87, 15.42), (2.5, 6, 15), "GP"),
        ],
    )
    def test_dual_classification(self, al, psd, dist, expected):
        uscs_clf = create_uscs_classifier(*al, *psd, *dist)
        assert uscs_clf.classify().symbol == expected

    @pytest.mark.parametrize(
        "al,psd,expected",
        [
            ((30.8, 20.7), (10.29, 81.89), "SW-SC,SP-SC"),
            ((24.4, 14.7), (9.77, 44.82), "GW-GC,GP-GC"),
            ((49.5, 33.6), (6.93, 91.79), "SW-SM,SP-SM"),
            ((30.33, 23.42), (8.93, 7.69), "GW-GM,GP-GM"),
            ((35.32, 25.57), (9.70, 5.63), "GW-GM,GP-GM"),
            ((26.17, 19.69), (12.00, 8.24), "GW-GC,GP-GC"),
            ((32.78, 22.99), (3.87, 15.42), "GW,GP"),
        ],
    )
    def test_dual_classification_no_psd_coeff(self, al, psd, expected):
        uscs_clf = create_uscs_classifier(*al, *psd)
        assert uscs_clf.classify().symbol == expected

    @pytest.mark.parametrize(
        ["al", "psd", "organic", "expected"],
        [
            ((34.1, 21.1), (47.88, 37.84), False, "SC"),
            ((27.5, 13.8), (54.23, 45.69), False, "CL"),
            ((27.7, 22.7), (18.95, 77.21), False, "SM-SC"),
            ((64.1, 29), (57.17, 42.58), False, "CH"),
            ((56, 32.4), (51.11, 46.87), False, "MH"),
            ((70, 38), (86, 7), False, "MH"),
            ((26.4, 19.4), (54.76, 45.24), False, "ML-CL"),
            ((33, 21), (30, 30), False, "GC"),
            ((34.46, 23.85), (18.09, 18.7), False, "GC"),
            ((45, 16), (59, 41), False, "CL"),
            ((55, 40), (85, 15), False, "MH"),
            ((49.93, 37.22), (49.4, 35.98), False, "SM"),
            ((42.77, 29.98), (27.6, 27.93), False, "GM"),
            ((35.83, 25.16), (68.94, 28.88), False, "ML"),
            ((27.7, 22.7), (18.95, 20.48), False, "GM-GC"),
            ((35.83, 25.16), (68.94, 28.88), True, "OL"),
            ((55.0, 40.0), (85.0, 15.0), True, "OH"),
        ],
    )
    def test_single_classification(self, al, psd, organic, expected):
        uscs_clf = create_uscs_classifier(*al, *psd, organic=organic)
        assert uscs_clf.classify().symbol == expected



import enum
import functools
from typing import Callable

from func_validator import ValidationError

from .math import *

__all__ = ["AbstractStrEnum", "ValidationError", "add_repr", "round_"]


class StrEnumMeta(enum.EnumMeta):
    def __contains__(cls, item) -> bool:
        if isinstance(item, (str, cls)):
            return str(item) in (member.value for member in cls)
        return NotImplemented

    def __repr__(cls) -> str:
        return str([member.value for member in cls])


class AbstractStrEnum(enum.StrEnum, metaclass=StrEnumMeta):
    """An abstract string enumeration class that inherits from StrEnum.

    This class can be used as a base class for creating string
    enumerations.
    """


def add_repr(cls):
    """A class decorator that adds a __repr__ method to the class."""

    def __repr__(self) -> str:
        inst_attrs = self.__dict__
        attrs = (f"{key.strip('_')}={val}" for key, val in inst_attrs.items())
        return f"{type(self).__name__}({', '.join(attrs)})"

    def __str__(self) -> str:
        return repr(self)

    cls.__repr__ = __repr__
    cls.__str__ = __str__

    return cls


def round_(ndigits: int) -> Callable:
    """A decorator that rounds the result of a callable to a specified
    number of decimal places.

    The returned value of the callable should support the `__round__`
    dunder method and should be a numeric value.

    TypeError is raised when `ndigits` is not an int.
    """

    if not isinstance(ndigits, int):
        raise TypeError("ndigits must be an int")

    def dec(fn) -> Callable[..., float]:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs) -> float:
            res = fn(*args, **kwargs)
            return round(res, ndigits=ndigits)

        return wrapper

    return dec



import unittest

from func_validator import ValidationError

from geolysis.foundation import (
    CircularFooting,
    Foundation,
    RectangularFooting,
    Shape,
    SquareFooting,
    StripFooting,
    create_foundation,
)


class TestCircularFooting(unittest.TestCase):
    def test_attributes(self):
        circ_footing = CircularFooting(diameter=1.2)
        circ_footing.width = 1.4
        self.assertAlmostEqual(circ_footing.length, 1.4)
        self.assertAlmostEqual(circ_footing.diameter, 1.4)

        circ_footing.length = 1.5
        self.assertAlmostEqual(circ_footing.width, 1.5)
        self.assertAlmostEqual(circ_footing.diameter, 1.5)


class TestSquareFooting(unittest.TestCase):
    def test_attributes(self):
        sqr_footing = SquareFooting(1.2)
        sqr_footing.length = 1.4
        self.assertAlmostEqual(sqr_footing.width, 1.4)


class TestFoundation(unittest.TestCase):

    def test_strip_footing(self):
        footing = StripFooting(width=2.0)
        self.assertEqual(footing.width, 2.0)
        self.assertEqual(footing.length, float("inf"))
        self.assertEqual(footing.shape, Shape.STRIP)

    def test_circular_footing(self):
        footing = CircularFooting(diameter=3.0)
        self.assertEqual(footing.diameter, 3.0)
        self.assertEqual(footing.width, 3.0)
        self.assertEqual(footing.length, 3.0)
        self.assertEqual(footing.shape, Shape.CIRCLE)

    def test_square_footing(self):
        footing = SquareFooting(width=2.5)
        self.assertEqual(footing.width, 2.5)
        self.assertEqual(footing.length, 2.5)
        self.assertEqual(footing.shape, Shape.SQUARE)

    def test_rectangular_footing(self):
        footing = RectangularFooting(width=2.0, length=3.0)
        self.assertEqual(footing.width, 2.0)
        self.assertEqual(footing.length, 3.0)
        self.assertEqual(footing.shape, Shape.RECTANGLE)

    def test_foundation_size(self):
        footing = StripFooting(width=2.0)
        foundation = Foundation(depth=1.5, footing_size=footing,
                                eccentricity=0.1)
        self.assertEqual(foundation.depth, 1.5)
        self.assertEqual(foundation.width, 2.0)
        self.assertEqual(foundation.length, float("inf"))
        self.assertEqual(foundation.eccentricity, 0.1)
        self.assertEqual(foundation.effective_width, 1.8)
        self.assertEqual(foundation.footing_shape, Shape.STRIP)

        foundation.width = 1.4
        foundation.length = 1.5
        foundation.ground_water_level = 1.8

        self.assertEqual(foundation.width, 1.4)
        self.assertEqual(foundation.length, 1.5)
        self.assertEqual(foundation.ground_water_level, 1.8)

    def test_create_foundation_strip(self):
        foundation = create_foundation(depth=1.5, width=2.0, shape=Shape.STRIP)
        self.assertEqual(foundation.depth, 1.5)
        self.assertEqual(foundation.width, 2.0)
        self.assertEqual(foundation.length, float("inf"))
        self.assertEqual(foundation.footing_shape, Shape.STRIP)

    def test_create_foundation_square(self):
        foundation = create_foundation(depth=1.5, width=2.5,
                                       shape=Shape.SQUARE)
        self.assertEqual(foundation.depth, 1.5)
        self.assertEqual(foundation.width, 2.5)
        self.assertEqual(foundation.length, 2.5)
        self.assertEqual(foundation.footing_shape, Shape.SQUARE)

    def test_create_foundation_circle(self):
        foundation = create_foundation(depth=1.5, width=3.0,
                                       shape=Shape.CIRCLE)
        self.assertEqual(foundation.depth, 1.5)
        self.assertEqual(foundation.width, 3.0)
        self.assertEqual(foundation.length, 3.0)
        self.assertEqual(foundation.footing_shape, Shape.CIRCLE)

    def test_create_foundation_rectangle(self):
        foundation = create_foundation(
            depth=1.5, width=2.0, length=3.0, shape=Shape.RECTANGLE
        )
        self.assertEqual(foundation.depth, 1.5)
        self.assertEqual(foundation.width, 2.0)
        self.assertEqual(foundation.length, 3.0)
        self.assertEqual(foundation.footing_shape, Shape.RECTANGLE)

    def test_create_foundation_invalid_shape(self):
        with self.assertRaises(ValidationError):
            create_foundation(depth=1.5, width=2.0, shape="invalid_shape")

    def test_create_foundation_missing_length(self):
        with self.assertRaises(ValueError):
            create_foundation(depth=1.5, width=2.0, shape=Shape.RECTANGLE)



from geolysis.foundation import Shape
from geolysis.utils import isclose, round_, sindeg, tandeg, atan, add_repr
from ._core import UltimateBearingCapacity
from ._hansen_ubc import HansenBearingCapacityFactors

__all__ = ["VesicUltimateBearingCapacity"]


class VesicBearingCapacityFactors:

    @staticmethod
    @round_(ndigits=2)
    def n_c(friction_angle: float) -> float:
        return HansenBearingCapacityFactors.n_c(friction_angle)

    @staticmethod
    @round_(ndigits=2)
    def n_q(friction_angle: float) -> float:
        return HansenBearingCapacityFactors.n_q(friction_angle)

    @staticmethod
    @round_(ndigits=2)
    def n_gamma(friction_angle: float) -> float:
        return (
            2.0
            * (VesicBearingCapacityFactors.n_q(friction_angle) + 1.0)
            * tandeg(friction_angle)
        )


class VesicShapeFactors:
    @staticmethod
    @round_(ndigits=3)
    def s_c(
        friction_angle: float,
        f_width: float,
        f_length: float,
        f_shape: Shape,
    ) -> float:
        n_q = VesicBearingCapacityFactors.n_q(friction_angle)
        n_c = VesicBearingCapacityFactors.n_c(friction_angle)

        if f_shape == Shape.STRIP:
            return 1.0
        else:
            return 1.0 + (f_width / f_length) * (n_q / n_c)

    @staticmethod
    @round_(ndigits=3)
    def s_q(
        friction_angle: float,
        f_width: float,
        f_length: float,
        f_shape: Shape,
    ) -> float:
        if f_shape == Shape.STRIP:
            return 1.0
        else:
            return 1.0 + (f_width / f_length) * tandeg(friction_angle)

    @staticmethod
    @round_(ndigits=3)
    def s_gamma(f_width: float, f_length: float, f_shape: Shape) -> float:
        if f_shape == Shape.STRIP:
            return 1.0
        else:
            return 1.0 - 0.4 * (f_width / f_length)


class VesicDepthFactors:

    @staticmethod
    def _d_c(friction_angle: float, f_depth: float, f_width: float) -> float:
        d_q = VesicDepthFactors.d_q(friction_angle, f_depth, f_width)
        n_c = VesicBearingCapacityFactors.n_c(friction_angle)
        return d_q - ((1.0 - d_q) / (n_c * tandeg(friction_angle)))

    @staticmethod
    @round_(ndigits=3)
    def d_c(friction_angle: float, f_depth: float, f_width: float) -> float:
        d2w = round(f_depth / f_width, 1)

        if d2w <= 1.0:
            if isclose(friction_angle, 0.0):
                _d_c = 1.0 + 0.4 * d2w
            else:
                _d_c = VesicDepthFactors._d_c(friction_angle, f_depth, f_width)
        else:
            if isclose(friction_angle, 0.0):
                _d_c = 1.0 + 0.4 * atan(d2w)
            else:
                _d_c = VesicDepthFactors._d_c(friction_angle, f_depth, f_width)

        return _d_c

    @staticmethod
    @round_(ndigits=3)
    def d_q(friction_angle: float, f_depth: float, f_width: float) -> float:
        d2w = f_depth / f_width

        if d2w <= 1.0:
            if isclose(friction_angle, 0.0):
                _d_q = 1.0
            else:
                _d_q = (
                    1.0
                    + 2
                    * tandeg(friction_angle)
                    * (1 - sindeg(friction_angle)) ** 2
                    * d2w
                )
        else:
            if isclose(friction_angle, 0.0):
                _d_q = 1.0
            else:
                _d_q = 1.0 + (
                    2.0 * tandeg(friction_angle) * (1 - sindeg(friction_angle)) ** 2
                ) * atan(d2w)
        return _d_q

    @staticmethod
    def d_gamma() -> float:
        return 1.0


class VesicInclinationFactors:

    @staticmethod
    @round_(ndigits=3)
    def i_c(load_angle: float) -> float:
        return (1.0 - load_angle / 90.0) ** 2.0

    @staticmethod
    @round_(ndigits=3)
    def i_q(load_angle: float) -> float:
        return VesicInclinationFactors.i_c(load_angle)

    @staticmethod
    @round_(ndigits=3)
    def i_gamma(friction_angle: float, load_angle: float) -> float:
        if isclose(friction_angle, 0.0):
            return 1.0
        return (1.0 - load_angle / friction_angle) ** 2.0


@add_repr
class VesicUltimateBearingCapacity(UltimateBearingCapacity):
    """Ultimate bearing capacity for soils according to `Vesic (1973)`.

    See [implementation](../formulas/ultimate-bearing-capacity.md/#vesic-bearing-capacity)
    for more details on bearing capacity equation used.
    """

    @property
    def n_c(self) -> float:
        r"""Bearing capacity factor $N_c$."""
        return VesicBearingCapacityFactors.n_c(self.friction_angle)

    @property
    def n_q(self) -> float:
        r"""Bearing capacity factor $N_q$."""
        return VesicBearingCapacityFactors.n_q(self.friction_angle)

    @property
    def n_gamma(self) -> float:
        r"""Bearing capacity factor $N_{\gamma}$."""
        return VesicBearingCapacityFactors.n_gamma(self.friction_angle)

    @property
    def s_c(self) -> float:
        r"""Shape factor $S_c$."""
        width, length, shape = self.foundation_size.footing_params()
        return VesicShapeFactors.s_c(self.friction_angle, width, length, shape)

    @property
    def s_q(self) -> float:
        r"""Shape factor $S_q$."""
        width, length, shape = self.foundation_size.footing_params()
        return VesicShapeFactors.s_q(self.friction_angle, width, length, shape)

    @property
    def s_gamma(self) -> float:
        r"""Shape factor $S_{\gamma}$."""
        width, length, shape = self.foundation_size.footing_params()
        return VesicShapeFactors.s_gamma(width, length, shape)

    @property
    def d_c(self) -> float:
        r"""Depth factor $D_c$."""
        depth, width = self.foundation_size.depth, self.foundation_size.width
        return VesicDepthFactors.d_c(self.friction_angle, depth, width)

    @property
    def d_q(self) -> float:
        r"""Depth factor $D_q$."""
        depth, width = self.foundation_size.depth, self.foundation_size.width
        return VesicDepthFactors.d_q(self.friction_angle, depth, width)

    @property
    def d_gamma(self) -> float:
        r"""Depth factor $D_{\gamma}$."""
        return VesicDepthFactors.d_gamma()

    @property
    def i_c(self) -> float:
        r"""Inclination factor $I_c$."""
        return VesicInclinationFactors.i_c(self.load_angle)

    @property
    def i_q(self) -> float:
        r"""Inclination factor $I_q$."""
        return VesicInclinationFactors.i_q(self.load_angle)

    @property
    def i_gamma(self) -> float:
        r"""Inclination factor $I_{\gamma}$."""
        return VesicInclinationFactors.i_gamma(self.friction_angle, self.load_angle)



from geolysis.foundation import Foundation
from geolysis.utils import round_

from ._core import AllowableBearingCapacity


class MeyerhofABC4PadFoundation(AllowableBearingCapacity):
    """Allowable bearing capacity for pad foundation on cohesionless
    soils according to `Meyerhof (1956)`.

    See [implementation](../formulas/allowable-bearing-capacity.md/#meyerhof-bearing-capacity-for-pad-foundation)
    for more details on bearing capacity equation used.

    """

    def __init__(
        self,
        corrected_spt_n_value: float,
        tol_settlement: float,
        foundation_size: Foundation,
    ):
        """
        :param corrected_spt_n_value: Average uncorrected SPT N-value
                                      (60% energy with dilatancy (water)
                                      correction if applicable) within
                                      the foundation influence zone i.e.
                                      $D_f$ to $D_f + 2B$.
        :param tol_settlement: Tolerable settlement of foundation (mm).
        :param foundation_size: Size of the foundation.
        """
        super().__init__(
            corrected_spt_n_value=corrected_spt_n_value,
            tol_settlement=tol_settlement,
            foundation_size=foundation_size,
        )

    @round_(ndigits=2)
    def _bearing_capacity(self):
        """
        Calculates the allowable bearing capacity of the pad foundation.
        """
        n_corr = self.corrected_spt_n_value
        width = self.foundation_size.width

        if width <= 1.2:
            return 12 * n_corr * self._fd() * self._sr()

        return (
            8
            * n_corr
            * ((3.28 * width + 1) / (3.28 * width)) ** 2
            * self._fd()
            * self._sr()
        )


class MeyerhofABC4MatFoundation(MeyerhofABC4PadFoundation):
    """Allowable bearing capacity for mat foundation on cohesionless
    soils according to `Meyerhof (1956)`.

    See [implementation](../formulas/allowable-bearing-capacity.md/#meyerhof-bearing-capacity-for-mat-foundation)
    for more details on bearing capacity equation used.

    """

    @round_(ndigits=2)
    def _bearing_capacity(self):
        """Calculate the allowable bearing capacity of the mat foundation."""
        n_corr = self.corrected_spt_n_value
        return 8 * n_corr * self._fd() * self._sr()



from geolysis.foundation import Shape
from geolysis.utils import (
    cosdeg,
    cotdeg,
    exp,
    isclose,
    pi,
    round_,
    sindeg,
    tandeg,
    add_repr,
)

from ._core import UltimateBearingCapacity

__all__ = ["HansenUltimateBearingCapacity"]


class HansenBearingCapacityFactors:

    @staticmethod
    @round_(ndigits=2)
    def n_c(friction_angle: float) -> float:
        if isclose(friction_angle, 0.0):
            return 5.14
        return cotdeg(friction_angle) * (
            HansenBearingCapacityFactors.n_q(friction_angle) - 1.0
        )

    @staticmethod
    @round_(ndigits=2)
    def n_q(friction_angle: float) -> float:
        return tandeg(45.0 + friction_angle / 2.0) ** 2.0 * exp(
            pi * tandeg(friction_angle)
        )

    @staticmethod
    @round_(ndigits=2)
    def n_gamma(friction_angle: float) -> float:
        return (
            1.8
            * (HansenBearingCapacityFactors.n_q(friction_angle) - 1.0)
            * tandeg(friction_angle)
        )


class HansenShapeFactors:

    @staticmethod
    @round_(ndigits=2)
    def s_c(f_width: float, f_length: float, f_shape: Shape) -> float:
        if f_shape == Shape.STRIP:
            return 1.0
        elif f_shape == Shape.RECTANGLE:
            return 1.0 + 0.2 * f_width / f_length
        else:  # SQUARE & CIRCLE
            return 1.3

    @staticmethod
    @round_(ndigits=2)
    def s_q(f_width: float, f_length: float, f_shape: Shape) -> float:
        if f_shape == Shape.STRIP:
            return 1.0
        elif f_shape == Shape.RECTANGLE:
            return 1.0 + 0.2 * f_width / f_length
        else:  # SQUARE & CIRCLE
            return 1.2

    @staticmethod
    @round_(ndigits=2)
    def s_gamma(f_width: float, f_length: float, f_shape: Shape) -> float:
        if f_shape == Shape.STRIP:
            return 1.0
        elif f_shape == Shape.RECTANGLE:
            return 1.0 - 0.4 * f_width / f_length
        elif f_shape == Shape.SQUARE:
            return 0.8
        else:  # CIRCLE
            return 0.6


class HansenDepthFactors:

    @staticmethod
    @round_(ndigits=2)
    def d_c(f_depth: float, f_width: float) -> float:
        return 1.0 + 0.35 * f_depth / f_width

    @staticmethod
    @round_(ndigits=2)
    def d_q(f_depth: float, f_width: float) -> float:
        return HansenDepthFactors.d_c(f_depth, f_width)

    @staticmethod
    @round_(ndigits=2)
    def d_gamma() -> float:
        return 1.0


class HansenInclinationFactors:

    @staticmethod
    @round_(ndigits=2)
    def i_c(
        cohesion: float,
        load_angle: float,
        f_width: float,
        f_length: float,
    ) -> float:
        return 1.0 - sindeg(load_angle) / (2.0 * cohesion * f_width * f_length)

    @staticmethod
    @round_(ndigits=2)
    def i_q(load_angle: float) -> float:
        return 1.0 - (1.5 * sindeg(load_angle)) / cosdeg(load_angle)

    @staticmethod
    @round_(ndigits=2)
    def i_gamma(load_angle: float) -> float:
        return HansenInclinationFactors.i_q(load_angle) ** 2.0


@add_repr
class HansenUltimateBearingCapacity(UltimateBearingCapacity):
    r"""Ultimate bearing capacity for soils according to `Hansen (1961)`.

    See [implementation](../formulas/ultimate-bearing-capacity.md/#hansen-bearing-capacity)
    for more details on bearing capacity equation used.

    """

    @property
    def n_c(self) -> float:
        r"""Bearing capacity factor $N_c$."""
        return HansenBearingCapacityFactors.n_c(self.friction_angle)

    @property
    def n_q(self) -> float:
        r"""Bearing capacity factor $N_q$."""
        return HansenBearingCapacityFactors.n_q(self.friction_angle)

    @property
    def n_gamma(self) -> float:
        r"""Bearing capacity factor $N_{\gamma}$."""
        return HansenBearingCapacityFactors.n_gamma(self.friction_angle)

    @property
    def s_c(self) -> float:
        r"""Shape factor $S_c$."""
        width, length, shape = self.foundation_size.footing_params()
        return HansenShapeFactors.s_c(width, length, shape)

    @property
    def s_q(self) -> float:
        r"""Shape factor $S_q$."""
        width, length, shape = self.foundation_size.footing_params()
        return HansenShapeFactors.s_q(width, length, shape)

    @property
    def s_gamma(self) -> float:
        r"""Shape factor $S_{\gamma}$."""
        width, length, shape = self.foundation_size.footing_params()
        return HansenShapeFactors.s_gamma(width, length, shape)

    @property
    def d_c(self) -> float:
        r"""Depth factor $D_c$."""
        depth = self.foundation_size.depth
        width = self.foundation_size.width
        return HansenDepthFactors.d_c(depth, width)

    @property
    def d_q(self) -> float:
        r"""Depth factor $D_q$."""
        depth = self.foundation_size.depth
        width = self.foundation_size.width
        return HansenDepthFactors.d_q(depth, width)

    @property
    def d_gamma(self) -> float:
        r"""Depth factor $D_{\gamma}$."""
        return HansenDepthFactors.d_gamma()

    @property
    def i_c(self) -> float:
        r"""Inclination factor $I_c$."""
        width, length = self.foundation_size.width, self.foundation_size.length
        return HansenInclinationFactors.i_c(
            self.cohesion, self.load_angle, width, length
        )

    @property
    def i_q(self) -> float:
        r"""Inclination factor $I_q$."""
        return HansenInclinationFactors.i_q(self.load_angle)

    @property
    def i_gamma(self) -> float:
        r"""Inclination factor $I_{\gamma}$."""
        return HansenInclinationFactors.i_gamma(self.load_angle)



import pytest
from func_validator import ValidationError

from geolysis.spt import (
    DilatancyCorrection,
    EnergyCorrection,
    HammerType,
    SamplerType,
    SPT,
    create_overburden_pressure_correction,
)


def test_create_spt_correction_errors():
    with pytest.raises(ValidationError):
        create_overburden_pressure_correction(
            std_spt_n_value=34, eop=100, opc_type="TERZAGHI"
        )


class TestSPTNDesign:

    def test_spt_n_design(self):
        spt_design = SPT([7.0, 15.0, 18], method="min")

        assert spt_design.n_design() == pytest.approx(7.0)
        spt_design.method = "avg"
        assert spt_design.n_design() == pytest.approx(13.3)
        spt_design.method = "wgt"
        assert spt_design.n_design() == pytest.approx(9.4)

    def test_errors(self):
        # Provided an empty value for corrected_spt_n_values
        with pytest.raises(ValidationError):
            SPT(corrected_spt_n_values=[])

        # Provided an invalid method
        with pytest.raises(ValidationError):
            SPT(corrected_spt_n_values=[7.0, 15.0, 18], method="max")

        # corrected_spt_n_values is greater than 100
        with pytest.raises(ValidationError):
            SPT(corrected_spt_n_values=[22, 44, 120])

        # corrected_spt_n_values is 0.0
        with pytest.raises(ValidationError):
            SPT(corrected_spt_n_values=[0.0, 15.0, 18])

        # corrected_spt_n_values is less than 0.0
        with pytest.raises(ValidationError):
            SPT(corrected_spt_n_values=[-10, 15.0, 18])


class TestEnergyCorrection:

    @pytest.mark.parametrize(
        [
            "rec_spt_n_val",
            "energy_percentage",
            "borehole_diameter",
            "rod_len",
            "hammer_type",
            "sampler_type",
            "expected",
        ],
        [
            (30, 0.6, 65.0, 3, HammerType.DONUT_1, SamplerType.STANDARD, 22.5),
            (30, 0.6, 120.0, 5, HammerType.AUTOMATIC, SamplerType.NON_STANDARD, 37.5),
            (30, 0.6, 170.0, 7, HammerType.DONUT_2, SamplerType.STANDARD, 27.3),
            (30, 0.6, 65.0, 12, HammerType.SAFETY, SamplerType.STANDARD, 27.5),
        ],
    )
    def test_energy_correction(
        self,
        rec_spt_n_val,
        energy_percentage,
        borehole_diameter,
        rod_len,
        hammer_type,
        sampler_type,
        expected,
    ):
        energy_corr = EnergyCorrection(
            recorded_spt_n_value=rec_spt_n_val,
            energy_percentage=energy_percentage,
            borehole_diameter=borehole_diameter,
            rod_length=rod_len,
            hammer_type=hammer_type,
            sampler_type=sampler_type,
        )

        assert energy_corr.standardized_spt_n_value() == pytest.approx(expected)

    def test_errors(self):
        # Provided an invalid value for hammer_type
        with pytest.raises(ValidationError):
            EnergyCorrection(recorded_spt_n_value=22, hammer_type="manual")

        # Provided an invalid value for sampler_type
        with pytest.raises(ValidationError):
            EnergyCorrection(recorded_spt_n_value=22, sampler_type="std")


class TestGibbsHoltzOPC:

    @pytest.mark.parametrize(
        ["std_spt_n_value", "eop", "expected"], [(22.5, 100.0, 23.2)]
    )
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="gibbs"
        )
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestBazaraaPeckOPC:

    @pytest.mark.parametrize(
        ["std_spt_n_value", "eop", "expected"],
        [(11.4, 54.8, 13.9), (22.5, 100.0, 21.0), (22.5, 71.8, 22.5)],
    )
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="bazaraa"
        )
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestPeckOPC:

    @pytest.mark.parametrize(
        ["std_spt_n_value", "eop", "expected"], [(22.5, 100.0, 22.5)]
    )
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="peck"
        )
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestLiaoWhitmanOPC:

    @pytest.mark.parametrize(
        ["std_spt_n_value", "eop", "expected"], [(22.5, 100.0, 22.5)]
    )
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="liao"
        )
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestSkemptonOPC:
    @pytest.mark.parametrize(
        ["std_spt_n_value", "eop", "expected"], [(22.5, 100.0, 22.0)]
    )
    def test_correction(self, std_spt_n_value, eop, expected):
        opc_corr = create_overburden_pressure_correction(
            std_spt_n_value=std_spt_n_value, eop=eop, opc_type="skempton"
        )
        assert opc_corr.corrected_spt_n_value() == pytest.approx(expected)


class TestDilatancyCorrection:

    @pytest.mark.parametrize(
        ["std_spt_n_value", "expected"], [(22.5, 18.8), (12.6, 12.6)]
    )
    def test_correction(self, std_spt_n_value, expected):
        corr = DilatancyCorrection(corr_spt_n_value=std_spt_n_value)
        assert corr.corrected_spt_n_value() == pytest.approx(expected)



import enum
from dataclasses import dataclass
from typing import Annotated, Sequence

from func_validator import validate_func_args, MustBeNonNegative

from .utils import isclose, round_

__all__ = [
    "AtterbergLimits",
    "PSD",
    "AASHTO",
    "USCS",
    "create_aashto_classifier",
    "create_uscs_classifier",
]


class SizeDistError(ZeroDivisionError):
    """Exception raised when size distribution is not provided."""


class _Clf(tuple, enum.Enum):

    def __str__(self) -> str:
        return self.name

    @property
    def symbol(self) -> str:
        return self.value[0]

    @property
    def description(self) -> str:
        return self.value[1]


class USCSSymbol(_Clf):
    """Unified Soil Classification System (USCS) symbols and descriptions.

    Each member represents a USCS soil type, grading, or plasticity symbol.
    Aliases are provided where applicable.
    """

    # General soil types
    G = ("G", "Gravel")
    """Gravel"""

    GRAVEL = G

    S = ("S", "Sand")
    """Sand"""

    SAND = S

    M = ("M", "Silt")
    """Silt"""

    SILT = M

    C = ("C", "Clay")
    """Clay"""

    CLAY = C

    O = ("O", "Organic")
    """Organic soil"""

    ORGANIC = O

    # Grading descriptors
    W = ("W", "Well graded")
    """Well graded"""

    WELL_GRADED = W

    P = ("P", "Poorly graded")
    """Poorly graded"""

    POORLY_GRADED = P

    # Plasticity descriptors
    L = ("L", "Low plasticity")
    """Low plasticity"""

    LOW_PLASTICITY = L

    H = ("H", "High plasticity")
    """High plasticity"""

    HIGH_PLASTICITY = H

    # Gravels
    GW = ("GW", "Well graded gravels")
    """Well graded gravels"""

    GP = ("GP", "Poorly graded gravels")
    """Poorly graded gravels"""

    GM = ("GM", "Silty gravels")
    """Silty gravels"""

    GC = ("GC", "Clayey gravels")
    """Clayey gravels"""

    GM_GC = ("GM-GC", "Gravelly clayey silt")
    """Gravelly clayey silt"""

    GW_GM = ("GW-GM", "Well graded gravel with silt")
    """Well graded gravel with silt"""

    GP_GM = ("GP-GM", "Poorly graded gravel with silt")
    """Poorly graded gravel with silt"""

    GW_GC = ("GW-GC", "Well graded gravel with clay")
    """Well graded gravel with clay"""

    GP_GC = ("GP-GC", "Poorly graded gravel with clay")
    """Poorly graded gravel with clay"""

    # Sands
    SW = ("SW", "Well graded sands")
    """Well graded sands"""

    SP = ("SP", "Poorly graded sands")
    """Poorly graded sands"""

    SM = ("SM", "Silty sands")
    """Silty sands"""

    SC = ("SC", "Clayey sands")
    """Clayey sands"""

    SM_SC = ("SM-SC", "Sandy clayey silt")
    """Sandy clayey silt"""

    SW_SM = ("SW-SM", "Well graded sand with silt")
    """Well graded sand with silt"""

    SP_SM = ("SP-SM", "Poorly graded sand with silt")
    """Poorly graded sand with silt"""

    SW_SC = ("SW-SC", "Well graded sand with clay")
    """Well graded sand with clay"""

    SP_SC = ("SP-SC", "Poorly graded sand with clay")
    """Poorly graded sand with clay"""

    # Silts and clays
    ML = ("ML", "Inorganic silts with low plasticity")
    """Inorganic silts with low plasticity"""

    CL = ("CL", "Inorganic clays with low plasticity")
    """Inorganic clays with low plasticity"""

    ML_CL = ("ML-CL", "Clayey silt with low plasticity")
    """Clayey silt with low plasticity"""

    OL = ("OL", "Organic clays with low plasticity")
    """Organic clays with low plasticity"""

    MH = ("MH", "Inorganic silts with high plasticity")
    """Inorganic silts with high plasticity"""

    CH = ("CH", "Inorganic clays with high plasticity")
    """Inorganic clays with high plasticity"""

    OH = ("OH", "Organic silts with high plasticity")
    """Organic silts with high plasticity"""

    Pt = ("Pt", "Highly organic soils")
    """Highly organic soils"""


class AASHTOSymbol(_Clf):
    """
    AASHTO soil classification symbols and descriptions.

    Each member represents a standard AASHTO soil class used in
    pavement and highway engineering.
    """

    A_1_a = ("A-1-a", "Stone fragments, gravel, and sand")
    """Stone fragments, gravel, and sand"""

    A_1_b = ("A-1-b", "Stone fragments, gravel, and sand")
    """Stone fragments, gravel, and sand"""

    A_3 = ("A-3", "Fine sand")
    """Fine sand"""

    A_2_4 = ("A-2-4", "Silty or clayey gravel and sand")
    """Silty or clayey gravel and sand"""

    A_2_5 = ("A-2-5", "Silty or clayey gravel and sand")
    """Silty or clayey gravel and sand"""

    A_2_6 = ("A-2-6", "Silty or clayey gravel and sand")
    """Silty or clayey gravel and sand"""

    A_2_7 = ("A-2-7", "Silty or clayey gravel and sand")
    """Silty or clayey gravel and sand"""

    A_4 = ("A-4", "Silty soils")
    """Silty soils"""

    A_5 = ("A-5", "Silty soils")
    """Silty soils"""

    A_6 = ("A-6", "Clayey soils")
    """Clayey soils"""

    A_7_5 = ("A-7-5", "Clayey soils")
    """Clayey soils"""

    A_7_6 = ("A-7-6", "Clayey soils")
    """Clayey soils"""


class AtterbergLimits:
    """Represents the water contents at which soil changes from one state
    to the other.
    """

    class __A_LINE:

        def __get__(self, obj, objtype=None) -> float:
            return 0.73 * (obj.liquid_limit - 20.0)

    A_LINE = __A_LINE()
    """The ``A-line`` determines if a soil is clayey or silty.
    
    $A = 0.73(LL - 20.0)$
    """

    def __init__(self, liquid_limit: float, plastic_limit: float):
        """
        :param liquid_limit: Water content beyond which soils flows
                             under their own weight (%). It can also be
                             defined as the minimum moisture content at
                             which a soil flows upon application of a
                             very small shear force.
        :param plastic_limit: Water content at which plastic deformation
                              can be initiated (%). It is also the
                              minimum water content at which soil can be
                              rolled into a thread 3mm thick (molded
                              without breaking).
        """
        self.liquid_limit = liquid_limit
        self.plastic_limit = plastic_limit

    @property
    def liquid_limit(self) -> float:
        """
        Water content beyond which soils flows under their own weight (%).
        """
        return self._liquid_limit

    @liquid_limit.setter
    @validate_func_args
    def liquid_limit(self, val: Annotated[float, MustBeNonNegative]):
        self._liquid_limit = val

    @property
    def plastic_limit(self) -> float:
        """
        Water content at which plastic deformation can be initiated (%).
        """
        return self._plastic_limit

    @plastic_limit.setter
    @validate_func_args
    def plastic_limit(self, val: Annotated[float, MustBeNonNegative]):
        if self.liquid_limit < val:
            msg = (
                f"plastic_limit: {val} cannot be greater than "
                f"liquid_limit: {self.liquid_limit}"
            )
            raise ValueError(msg)

        self._plastic_limit = val

    @property
    @round_(2)
    def plasticity_index(self) -> float:
        """Plasticity index (PI) is the range of water content over which
        the soil remains in the plastic state.

        It is also the numerical difference between the liquid limit and
        plastic limit of the soil.

        $$PI = LL - PL$$
        """
        return self.liquid_limit - self.plastic_limit

    @property
    def fine_material_type(self) -> USCSSymbol:
        """Checks whether the soil is either clay or silt."""
        return USCSSymbol.CLAY if self.above_A_LINE() else USCSSymbol.SILT

    def above_A_LINE(self) -> bool:
        """Checks if the soil sample is above A-Line."""
        return self.plasticity_index > self.A_LINE

    def limit_plot_in_hatched_zone(self) -> bool:
        """Checks if soil sample plot in the hatched zone on the
        atterberg chart.
        """
        return 4 <= self.plasticity_index <= 7 and 10 < self.liquid_limit < 30

    @round_(ndigits=2)
    def liquidity_index(self, nmc: float) -> float:
        r"""Return the liquidity index of the soil.

        $$I_l = \dfrac{w - PL}{PI} \cdot 100$$

        Liquidity index of a soil indicates the nearness of its
        `natural water content` to its `liquid limit`. When the soil
        is at the plastic limit its liquidity index is zero. Negative
        values of the liquidity index indicate that the soil is in a
        hard (desiccated) state. It is also known as Water-Plasticity
        ratio.

        :param nmc: Moisture contents of the soil in natural condition.
        """
        return ((nmc - self.plastic_limit) / self.plasticity_index) * 100.0

    @round_(2)
    def consistency_index(self, nmc: float) -> float:
        r"""Return the consistency index of the soil.

        $$I_c = \dfrac{LL - w}{PI} \cdot 100$$

        Consistency index indicates the consistency (firmness) of soil.
        It shows the nearness of the ``natural water content`` of the
        soil to its `plastic limit`. When the soil is at the liquid
        limit, the consistency index is zero. The soil at consistency
        index of zero will be extremely soft and has negligible shear
        strength. A soil at a water content equal to the plastic limit
        has consistency index of 100% indicating that the soil is
        relatively firm. A consistency index of greater than 100% shows
        the soil is relatively strong (semi-solid state). A negative
        value indicate the soil is in the liquid state. It is also known
        as Relative Consistency.

        :param nmc: Moisture contents of the soil in natural condition.
        """
        return ((self.liquid_limit - nmc) / self.plasticity_index) * 100.0


class _SizeDistribution:
    """Particle size distribution of soil sample.

    Features obtained from the Particle Size Distribution graph.
    """

    def __init__(self, d_10: float = 0.0, d_30: float = 0.0, d_60: float = 0.0):
        self.d_10 = d_10
        self.d_30 = d_30
        self.d_60 = d_60

    def __iter__(self):
        return iter([self.d_10, self.d_30, self.d_60])

    @property
    def coeff_of_curvature(self) -> float:
        return (self.d_30**2.0) / (self.d_60 * self.d_10)

    @property
    def coeff_of_uniformity(self) -> float:
        return self.d_60 / self.d_10

    def grade(self, coarse_soil: USCSSymbol) -> USCSSymbol:
        """Grade of soil sample. Soil grade can either be well graded or
        poorly graded.

        :param coarse_soil: Coarse fraction of the soil sample.
        """
        if coarse_soil is USCSSymbol.GRAVEL:
            if 1 < self.coeff_of_curvature < 3 and self.coeff_of_uniformity >= 4:
                grade = USCSSymbol.WELL_GRADED
            else:
                grade = USCSSymbol.POORLY_GRADED
            return grade

        # coarse soil is sand
        if 1 < self.coeff_of_curvature < 3 and self.coeff_of_uniformity >= 6:
            grade = USCSSymbol.WELL_GRADED
        else:
            grade = USCSSymbol.POORLY_GRADED
        return grade


class PSD:
    """Quantitative proportions by mass of various sizes of particles
    present in a soil.
    """

    def __init__(
        self,
        fines: float,
        sand: float,
        d_10: float = 0,
        d_30: float = 0,
        d_60: float = 0,
    ):
        """
        :param fines: Percentage of fines in soil sample (%) i.e. The
                      percentage of soil sample passing through No. 200
                      sieve (0.075mm).
        :param sand: Percentage of sand in soil sample (%).
        :param d_10: Diameter at which 10% of the soil by weight is finer.
        :param d_30: Diameter at which 30% of the soil by weight is finer.
        :param d_60: Diameter at which 60% of the soil by weight is finer.
        """
        self.fines = fines
        self.sand = sand
        self.size_dist = _SizeDistribution(d_10=d_10, d_30=d_30, d_60=d_60)

    @property
    def gravel(self):
        """Percentage of gravel in soil sample (%)."""
        return 100.0 - (self.fines + self.sand)

    @property
    def coarse_material_type(self) -> USCSSymbol:
        """Determines whether the soil is either gravel or sand."""
        if self.gravel > self.sand:
            return USCSSymbol.GRAVEL
        return USCSSymbol.SAND

    @property
    @round_(ndigits=2)
    def coeff_of_curvature(self) -> float:
        r"""Coefficient of curvature of soil sample.

        $$C_c = \dfrac{D^2_{30}}{D_{60} \cdot D_{10}}$$

        For the soil to be well graded, the value of $C_c$ must be
        between 1 and 3.
        """
        return self.size_dist.coeff_of_curvature

    @property
    @round_(ndigits=2)
    def coeff_of_uniformity(self) -> float:
        r"""Coefficient of uniformity of soil sample.

        $$C_u = \dfrac{D_{60}}{D_{10}}$$

        $C_u$ value greater than 4 to 6 classifies the soil as well
        graded for gravels and sands respectively. When $C_u$ is less
        than 4, it is classified as poorly graded or uniformly graded
        soil.

        Higher values of $C_u$ indicates that the soil mass consists of
        soil particles with different size ranges.
        """
        return self.size_dist.coeff_of_uniformity

    def has_particle_sizes(self) -> bool:
        """Checks if soil sample has particle sizes."""
        return any(self.size_dist)

    def grade(self) -> USCSSymbol:
        r"""Return the grade of the soil sample, either well graded or
        poorly graded.

        Conditions for a well-graded soil:

        - $1 \lt C_c \lt 3$ and $C_u \ge 4$ (for gravels)
        - $1 \lt C_c \lt 3$ and $C_u \ge 6$ (for sands)
        """
        return self.size_dist.grade(coarse_soil=self.coarse_material_type)


@dataclass
class AASHTOResult:
    symbol: str
    symbol_no_group_idx: str
    description: str
    group_index: str


class AASHTO:
    r"""American Association of State Highway and Transportation
    Officials (AASHTO) classification system.

    The AASHTO classification system is useful for classifying soils for
    highways. It categorizes soils for highways based on particle size
    analysis and plasticity characteristics. It classifies both
    coarse-grained and fine-grained soils into eight main groups (A1-A7)
    with subgroups, along with a separate category (A8) for organic
    soils.

    - `A1 ~ A3` (Granular Materials) $\le$ 35% pass No. 200 sieve
    - `A4 ~ A7` (Silt-clay Materials) $\ge$ 36% pass No. 200 sieve
    - `A8` (Organic Materials)

    The Group Index `(GI)` is used to further evaluate soils within a
    group.

    !!! note

        The `GI` must be mentioned even when it is zero, to indicate
        that the soil has been classified as per AASHTO system.


    $$
    GI = (F_{200} - 35)[0.2 + 0.005(LL - 40)] + 0.01(F_{200} - 15)(PI - 10)
    $$
    """

    def __init__(self, atterberg_limits: AtterbergLimits, fines: float):
        """
        :param atterberg_limits: Atterberg limits of soil sample.

        :param fines: Percentage of fines in soil sample (%) i.e. The
                      percentage of soil sample passing through No. 200
                      sieve (0.075mm).
        """
        self.atterberg_limits = atterberg_limits
        self.fines = fines

    @property
    def fines(self) -> float:
        """Percentage of fines in soil sample (%)."""
        return self._fines

    @fines.setter
    @validate_func_args
    def fines(self, val: Annotated[float, MustBeNonNegative]):
        self._fines = val

    @round_(ndigits=0)
    def group_index(self) -> float:
        """Return the Group Index (GI) of the soil sample."""
        liquid_lmt = self.atterberg_limits.liquid_limit
        plasticity_idx = self.atterberg_limits.plasticity_index
        fines = self.fines

        x_1 = 1.0 if (x_0 := fines - 35.0) < 0.0 else min(x_0, 40.0)
        x_2 = 1.0 if (x_0 := liquid_lmt - 40.0) < 0.0 else min(x_0, 20.0)
        x_3 = 1.0 if (x_0 := fines - 15.0) < 0.0 else min(x_0, 40.0)
        x_4 = 1.0 if (x_0 := plasticity_idx - 10.0) < 0.0 else min(x_0, 20.0)

        return x_1 * (0.2 + 0.005 * x_2) + 0.01 * x_3 * x_4

    def classify(self) -> AASHTOResult:
        """Return the AASHTO classification of the soil."""
        soil_clf = self._classify()

        symbol_no_grp_idx, description = soil_clf.symbol, soil_clf.description
        group_idx = f"{self.group_index():.0f}"
        symbol = f"{symbol_no_grp_idx}({group_idx})"

        return AASHTOResult(
            symbol=symbol,
            symbol_no_group_idx=symbol_no_grp_idx,
            description=description,
            group_index=group_idx,
        )

    def _classify(self) -> AASHTOSymbol:
        # Silts A4-A7
        if self.fines > 35:
            soil_clf = self._fine_soil_classifier()
        # Coarse A1-A3
        else:
            soil_clf = self._coarse_soil_classifier()

        return soil_clf

    def _fine_soil_classifier(self) -> AASHTOSymbol:
        # A-4 -> A-5, Silty Soils
        # A-6 -> A-7, Clayey Soils
        liquid_lmt = self.atterberg_limits.liquid_limit
        plasticity_idx = self.atterberg_limits.plasticity_index

        if liquid_lmt <= 40:
            if plasticity_idx <= 10.0:
                soil_clf = AASHTOSymbol.A_4
            else:
                soil_clf = AASHTOSymbol.A_6
        else:
            if plasticity_idx <= 10.0:
                soil_clf = AASHTOSymbol.A_5
            else:
                if plasticity_idx <= (liquid_lmt - 30.0):
                    soil_clf = AASHTOSymbol.A_7_5
                else:
                    soil_clf = AASHTOSymbol.A_7_6

        return soil_clf

    def _coarse_soil_classifier(self) -> AASHTOSymbol:
        # A-3, Fine sand
        liquid_lmt = self.atterberg_limits.liquid_limit
        plasticity_idx = self.atterberg_limits.plasticity_index

        if self.fines <= 10.0 and isclose(plasticity_idx, 0.0, rel_tol=0.01):
            soil_clf = AASHTOSymbol.A_3
        # A-1-a -> A-1-b, Stone fragments, gravel, and sand
        elif self.fines <= 15 and plasticity_idx <= 6:
            soil_clf = AASHTOSymbol.A_1_a
        elif self.fines <= 25 and plasticity_idx <= 6:
            soil_clf = AASHTOSymbol.A_1_b
        # A-2-4 -> A-2-7, Silty or clayey gravel and sand
        elif liquid_lmt <= 40:
            if plasticity_idx <= 10:
                soil_clf = AASHTOSymbol.A_2_4
            else:
                soil_clf = AASHTOSymbol.A_2_6
        else:
            if plasticity_idx <= 10:
                soil_clf = AASHTOSymbol.A_2_5
            else:
                soil_clf = AASHTOSymbol.A_2_7

        return soil_clf


@dataclass
class USCSResult:
    symbol: str
    description: str


class USCS:
    """Unified Soil Classification System (USCS).

    The Unified Soil Classification System, initially developed by
    Casagrande in 1948 and later modified in 1952, is widely utilized in
    engineering projects involving soils. It is the most popular system
    for soil classification and is similar to Casagrande's
    Classification System. The system relies on particle size analysis
    and atterberg limits for classification.

    In this system, soils are first classified into two categories:

    - Coarse grained soils: If more than 50% of the soils is retained on
      No. 200 (0.075 mm) sieve, it is designated as coarse-grained soil.

    - Fine grained soils: If more than 50% of the soil passes through
      No. 200 sieve, it is designated as fine-grained soil.

    Highly Organic soils are identified by visual inspection. These
    soils are termed as Peat ($P_t$).
    """

    def __init__(
        self,
        atterberg_limits: AtterbergLimits,
        psd: PSD,
        organic=False,
    ):
        """
        :param atterberg_limits: Atterberg limits of the soil.
        :param psd: Particle size distribution of the soil.
        :param organic: Indicates whether soil is organic or not.
        """
        self.atterberg_limits = atterberg_limits
        self.psd = psd
        self.organic = organic

    def classify(self):
        """Return the USCS classification of the soil."""
        soil_clf = self._classify()

        # Ensure soil_clf is of type USCSSymbol
        if isinstance(soil_clf, str):
            soil_clf = USCSSymbol[soil_clf]

        if isinstance(soil_clf, USCSSymbol):
            soil_clf = USCSResult(
                symbol=soil_clf.symbol, description=soil_clf.description
            )
        else:
            # Handling tuple or list case for dual classification
            first_clf, second_clf = map(lambda clf: USCSSymbol[clf], soil_clf)
            comb_symbol = f"{first_clf.symbol},{second_clf.symbol}"
            comb_desc = f"{first_clf.description},{second_clf.description}"
            soil_clf = USCSResult(symbol=comb_symbol, description=comb_desc)

        return soil_clf

    def _classify(self) -> USCSSymbol | str | Sequence[str]:
        # Fine-grained, Run Atterberg
        if self.psd.fines > 50.0:
            return self._fine_soil_classifier()
        # Coarse grained, Run Sieve Analysis
        # Gravel or Sand
        return self._coarse_soil_classifier()

    def _fine_soil_classifier(self) -> USCSSymbol:
        liquid_lmt = self.atterberg_limits.liquid_limit
        plasticity_idx = self.atterberg_limits.plasticity_index

        if liquid_lmt < 50.0:
            # Low LL
            # Above A-line and PI > 7
            if self.atterberg_limits.above_A_LINE() and plasticity_idx > 7.0:
                soil_clf = USCSSymbol.CL
            # Limit plot in hatched area on plasticity chart
            elif self.atterberg_limits.limit_plot_in_hatched_zone():
                soil_clf = USCSSymbol.ML_CL
            # Below A-line or PI < 4
            else:
                soil_clf = USCSSymbol.OL if self.organic else USCSSymbol.ML
        # High LL
        else:
            # Above A-Line
            if self.atterberg_limits.above_A_LINE():
                soil_clf = USCSSymbol.CH
            # Below A-Line
            else:
                soil_clf = USCSSymbol.OH if self.organic else USCSSymbol.MH

        return soil_clf

    def _coarse_soil_classifier(self) -> USCSSymbol | str | Sequence[str]:
        coarse_material_type = self.psd.coarse_material_type

        # More than 12% pass No. 200 sieve
        if self.psd.fines > 12.0:
            # Above A-line
            if self.atterberg_limits.above_A_LINE():
                soil_clf = f"{coarse_material_type}{USCSSymbol.CLAY}"

            # Limit plot in hatched zone on plasticity chart
            elif self.atterberg_limits.limit_plot_in_hatched_zone():
                if coarse_material_type == USCSSymbol.GRAVEL:
                    soil_clf = USCSSymbol.GM_GC
                else:
                    soil_clf = USCSSymbol.SM_SC

            # Below A-line
            else:
                if coarse_material_type == USCSSymbol.GRAVEL:
                    soil_clf = USCSSymbol.GM
                else:
                    soil_clf = USCSSymbol.SM

        elif 5.0 <= self.psd.fines <= 12.0:
            # Requires dual symbol based on graduation and plasticity chart
            if self.psd.has_particle_sizes():
                soil_clf = self._dual_soil_classifier()
            else:
                fine_material_type = self.atterberg_limits.fine_material_type
                soil_clf = (
                    f"{coarse_material_type}{USCSSymbol.WELL_GRADED}_"
                    f"{coarse_material_type}{fine_material_type}",
                    f"{coarse_material_type}{USCSSymbol.POORLY_GRADED}_"
                    f"{coarse_material_type}{fine_material_type}",
                )

        # Less than 5% pass No. 200 sieve
        # Obtain Cc and Cu from grain size graph
        else:
            if self.psd.has_particle_sizes():
                soil_clf = f"{coarse_material_type}{self.psd.grade()}"

            else:
                soil_clf = (
                    f"{coarse_material_type}{USCSSymbol.WELL_GRADED}",
                    f"{coarse_material_type}{USCSSymbol.POORLY_GRADED}",
                )

        return soil_clf

    def _dual_soil_classifier(self) -> str:
        fine_material_type = self.atterberg_limits.fine_material_type
        coarse_material_type = self.psd.coarse_material_type

        return (
            f"{coarse_material_type}{self.psd.grade()}_"
            f"{coarse_material_type}{fine_material_type}"
        )


def create_aashto_classifier(
    liquid_limit: float,
    plastic_limit: float,
    fines: float,
) -> AASHTO:
    """A helper function that encapsulates the creation of a AASHTO
    classifier.

    :param liquid_limit: Water content beyond which soils flows under
                         their own weight (%). It can also be defined as
                         the minimum moisture content at which a soil
                         flows upon application of a very small shear
                         force.

    :param plastic_limit: Water content at which plastic deformation can
                          be initiated (%). It is also the minimum water
                          content at which soil can be rolled into a
                          thread 3mm thick (molded without breaking).

    :param fines: Percentage of fines in soil sample (%) i.e. The
                  percentage of soil sample passing through No. 200
                  sieve (0.075mm).
    """
    atterberg_limits = AtterbergLimits(liquid_limit, plastic_limit)
    return AASHTO(atterberg_limits, fines)


def create_uscs_classifier(
    liquid_limit: float,
    plastic_limit: float,
    fines: float,
    sand: float,
    d_10: float = 0,
    d_30: float = 0,
    d_60: float = 0,
    organic: bool = False,
):
    """A helper function that encapsulates the creation of a USCS
    classifier.

    :param liquid_limit: Water content beyond which soils flows under
                         their own weight (%). It can also be defined as
                         the minimum moisture content at which a soil
                         flows upon application of a very small shear
                         force.

    :param plastic_limit: Water content at which plastic deformation can
                          be initiated (%). It is also the minimum water
                          content at which soil can be rolled into a
                          thread 3mm thick. (molded without breaking)

    :param fines: Percentage of fines in soil sample (%) i.e. The
                  percentage of soil sample passing through No. 200
                  sieve (0.075mm).

    :param sand: Percentage of sand in soil sample (%).

    :param d_10: Diameter at which 10% of the soil by weight is finer.

    :param d_30: Diameter at which 30% of the soil by weight is finer.

    :param d_60: Diameter at which 60% of the soil by weight is finer.

    :param organic: Indicates whether soil is organic or not.
    """
    atterberg_lmts = AtterbergLimits(liquid_limit, plastic_limit)
    psd = PSD(fines=fines, sand=sand, d_10=d_10, d_30=d_30, d_60=d_60)
    return USCS(atterberg_limits=atterberg_lmts, psd=psd, organic=organic)
